import logging
from typing import Dict
from uuid import uuid4

from telegram import Update
from telegram.ext import CallbackContext, CommandHandler, Dispatcher, Filters, MessageHandler, Updater

from srai_telegrambot.command.command_base import CommandBase
from srai_telegrambot.command.command_image_tag import get_image_tag
from srai_telegrambot.dao_telegram_bot import DaoTelegramBot
from srai_telegrambot.mode.text_mode_base import TextModeBase
from srai_telegrambot.mode.voice_mode_base import VoiceModeBase
from srai_telegrambot.model.chat_message import ChatMessage
from srai_telegrambot.telegram_bot_interface import TelegramBotInterface

logger = logging.getLogger(__name__)
from threading import Thread


class TelegramBot(TelegramBotInterface):
    def __init__(
        self,
        token: str,
        dao_telegram_bot: DaoTelegramBot,
    ):
        self.token = token
        self.dao_telegram_bot = dao_telegram_bot
        self.dispatcher: Dispatcher = None  # type: ignore
        self.list_admin_ids = []
        self.dict_text_mode: Dict[str, TextModeBase] = {}
        self.text_mode_default: TextModeBase = None  # type: ignore
        self.dict_voice_mode: Dict[str, VoiceModeBase] = {}
        self.voice_mode_default: VoiceModeBase = None  # type: ignore
        self.dict_command: Dict[str, CommandBase] = {}
        self.thread: Thread = Thread(target=self._start)
        self.is_running = False
        # Create the Updater and pass it your bot's token.
        # Make sure to set use_context=True to use the new context based callbacks
        # Post version 12 this will no longer be necessary
        self.updater = Updater(self.token, use_context=True)
        self.dispatcher = self.updater.dispatcher  # type: ignore
        try:
            # on different commands - answer in Telegram

            # add text handler
            self.dispatcher.add_handler(MessageHandler(Filters.text & ~Filters.command, self.handle_text))
            self.dispatcher.add_handler(MessageHandler(Filters.voice & ~Filters.command, self.handle_voice))

            # log all errors
            self.dispatcher.add_error_handler(self.error)  # type: ignore
        except Exception as e:
            self.message_admins(f"Error during startup with image tag {get_image_tag()}: {e}")
            raise e

    def get_dao(self) -> DaoTelegramBot:
        return self.dao_telegram_bot

    def register_admin(self, admin_id: str):
        self.list_admin_ids.append(admin_id)

    def register_text_mode(self, text_mode: TextModeBase, set_default: bool = False):
        if text_mode.text_mode_name in self.dict_text_mode:
            raise Exception(f"Mode name {text_mode.text_mode_name} already registered")
        self.dict_text_mode[text_mode.text_mode_name] = text_mode
        text_mode.register(self)
        if set_default:
            self.text_mode_default = text_mode

    def register_voice_mode(self, voice_mode: VoiceModeBase, set_default: bool = False):
        if voice_mode.voice_mode_name in self.dict_voice_mode:
            raise Exception(f"Mode name {voice_mode.voice_mode_name} already registered")
        self.dict_voice_mode[voice_mode.voice_mode_name] = voice_mode
        voice_mode.register(self)
        if set_default:
            self.voice_mode_default = voice_mode

    def register_command(self, command: CommandBase):
        if command.command_name in self.dict_command:
            raise Exception(f"Command name {command.command_name} already registered")
        self.dict_command[command.command_name] = command
        command.register(self)
        self.dispatcher.add_handler(CommandHandler(command.command_name, command.execute_command_callback))

    def handle_text(self, update: Update, context: CallbackContext):
        """Handle text messages."""
        message_id = str(update.message.message_id)
        chat_id = str(update.message.chat_id)
        author_id = str(update.message.from_user.id)
        author_name = update.message.from_user.username
        message_content = update.message.to_dict()
        message = ChatMessage(message_id, chat_id, author_id, author_name, message_content)
        self.dao_telegram_bot.save_message(message)
        # TODO also have this catch the bot messages
        if self.text_mode_default is not None:
            self.text_mode_default.handle_text(update, context)

    def handle_voice(self, update: Update, context: CallbackContext):
        """Handle text messages."""
        message_id = str(update.message.message_id)
        chat_id = str(update.message.chat_id)
        author_id = str(update.message.from_user.id)
        author_name = update.message.from_user.username
        message_content = update.message.to_dict()
        message = ChatMessage(message_id, chat_id, author_id, author_name, message_content)
        self.dao_telegram_bot.save_message(message)
        # TODO also have this catch the bot messages
        if self.voice_mode_default is not None:
            self.voice_mode_default.handle_voice(update, context)

    def message_admins(self, text: str):
        for admin_id in self.list_admin_ids:
            self.message_chat(chat_id=admin_id, text=text)

    def message_chat(self, chat_id: str, text: str):
        self.updater.bot.send_message(chat_id=chat_id, text=text)
        message_id = str(uuid4())
        message_content = {"message_content_type": "text", "text": text}
        self.dao_telegram_bot.save_message(ChatMessage(message_id, str(chat_id), "0", "bot", message_content))

    def error(self, update: Update, context: CallbackContext):
        """Log Errors caused by Updates."""
        logger.warning('Update "%s" caused error "%s"', update, context.error)

    def start(self):
        """Start the bot."""
        self.is_running = True
        self.thread.start()

    def _start(self):
        """Start the bot."""

        self.message_admins(f"Startup succes with image tag {get_image_tag()}")
        while self.is_running:

            #  Start the Bot
            self.updater.start_polling()

            # Run the bot until you press Ctrl-C or the process receives SIGINT,
            # SIGTERM or SIGABRT. This should be used most of the time, since
            # start_polling() is non-blocking and will stop the bot gracefully.
            # self.updater.idle()



from typing import Dict, Optional

from srai_core.store.database_base import DatabaseBase
from srai_core.store.document_store_base import DocumentStoreBase
from srai_core.store.document_store_cached import DocumentStoreCached

from srai_telegrambot.model.chat_message import ChatMessage


class DaoTelegramBot:

    def __init__(self, database: DatabaseBase, cache_database=False) -> None:
        self.database = database
        self.list_document_store_name = [
            "skill_state",
            "mode_state",
            "chat_message",
        ]
        self.dict_document_store: Dict[str, DocumentStoreBase] = {}
        for name_store in self.list_document_store_name:
            if cache_database:
                self.dict_document_store[name_store] = DocumentStoreCached(database.get_document_store(name_store))
            else:
                self.dict_document_store[name_store] = database.get_document_store(name_store)

        self.store_skill_state = self.dict_document_store["skill_state"]
        self.store_mode_state = self.dict_document_store["mode_state"]
        self.store_chat_message = self.dict_document_store["chat_message"]

    def save_message(self, chat_message: ChatMessage) -> None:
        self.store_chat_message.save_document(chat_message.message_id, chat_message.to_dict())

    def save_skill_state(self, skill_state_id: str, skill_state: dict) -> None:
        self.store_skill_state.save_document(skill_state_id, skill_state)

    def load_skill_state(self, skill_state_id: str) -> dict:
        return self.store_skill_state.load_document(skill_state_id)

    def try_load_mode_state(self, mode_state_id: str) -> Optional[dict]:
        if not self.store_mode_state.exists_document(mode_state_id):
            return None
        return self.store_mode_state.load_document(mode_state_id)

    def save_mode_state(self, mode_state_id: str, mode_state: dict) -> None:
        self.store_mode_state.save_document(mode_state_id, mode_state)



import os

from srai_telegrambot.command.command_base import CommandBase


def get_image_tag():
    image_tag = os.environ.get("IMAGE_TAG")
    if image_tag is None:
        message = "IMAGE_TAG not set"
    else:
        message = f"{image_tag}"
    return message


class CommandImageTag(CommandBase):
    def __init__(self):
        super().__init__("image_tag")

    def execute_command(self, chat_id: str, command_message: str) -> None:
        """Send a message when the command /help is issued."""
        image_tag = get_image_tag()
        self.telegram_bot.message_chat(chat_id, f"image tag running:  {image_tag}")



class ChatMessage:
    def __init__(
        self,
        message_id: str,
        chat_id: str,
        author_id: str,
        author_name: str,
        message_content: dict,
    ):
        self.message_id = message_id
        self.chat_id = chat_id
        self.author_id = author_id
        self.author_name = author_name
        self.message_content = message_content

    def to_dict(self) -> dict:
        return {
            "message_id": self.message_id,
            "chat_id": self.chat_id,
            "author_id": self.author_id,
            "author_name": self.author_name,
            "message_content": self.message_content,
        }

    @staticmethod
    def from_dict(dict_message: dict) -> "ChatMessage":
        message_id = dict_message["message_id"]
        chat_id = dict_message["chat_id"]
        author_id = dict_message["author_id"]
        author_name = dict_message["author_name"]
        message_content = dict_message["message_content"]
        return ChatMessage(message_id, chat_id, author_id, author_name, message_content)



from srai_telegrambot.mode.text_mode_base import TextModeBase


class ModeDefault(TextModeBase):
    def __init__(self):
        super().__init__("text_mode_default")

    def handle_text(self, update, context):
        update.message.reply_text(
            "Default mode is active and not specific text response will be given to non command messages."
        )



import base64
import os
import pickle as pkl

from langchain.chains import ConversationalRetrievalChain
from langchain.memory import ConversationBufferMemory
from langchain.text_splitter import CharacterTextSplitter
from langchain_community.document_loaders import PyPDFLoader, TextLoader
from langchain_community.vectorstores import VectorStore

# from langchain_community.vectorstores import FAISS
from langchain_openai import ChatOpenAI, OpenAIEmbeddings

from srai_telegrambot.mode.text_mode_base import TextModeBase


class TextModeRag(TextModeBase):

    def __init__(self, path_dir_vectorstore: str):
        super().__init__("text_mode_rag")
        self.path_dir_vectorstore = path_dir_vectorstore
        if not os.path.exists(self.path_dir_vectorstore):
            os.makedirs(self.path_dir_vectorstore)

        self.list_path_file_pdf = []
        self.list_path_file_txt = []
        self.text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=200)

        # Create vector store
        embeddings = OpenAIEmbeddings()
        self.vectorstore: VectorStore = None  # type: ignore
        if not os.path.isfile(os.path.join(path_dir_vectorstore, "index.pkl")):
            self.rebuild_vectorstore()
        else:
            raise Exception("FAISS is not supported in this version of the library")
            # self.vectorstore = FAISS.load_local(
            #     path_dir_vectorstore,
            #     embeddings=embeddings,
            #     allow_dangerous_deserialization=True,
            # )
        # Create conversation chain
        self.llm = ChatOpenAI(temperature=0.7, model="gpt-4")

    def rebuild_vectorstore(self):
        list_page = []
        for path_file in self.list_path_file_pdf:
            loader = PyPDFLoader(file_path=path_file)
            list_page.extend(loader.load_and_split(self.text_splitter))
        for path_file in self.list_path_file_txt:
            loader = TextLoader(file_path=path_file)
            list_page.extend(loader.load_and_split(self.text_splitter))

        # self.vectorstore = FAISS.from_documents(list_page, embedding=OpenAIEmbeddings())
        # self.vectorstore.save_local(self.path_dir_vectorstore)
        raise Exception("FAISS is not supported in this version of the library")

    def add_path_file_pdf(self, path_file: str):
        self.list_path_file_pdf.append(path_file)

    def add_path_file_txt(self, path_file: str):
        self.list_path_file_txt.append(path_file)

    def _handle_text(self, chat_id: str, prompt: str) -> str:
        mode_state = self.try_load_mode_state(chat_id)
        if mode_state is None:
            memory = ConversationBufferMemory(memory_key="chat_history", return_messages=True)
        else:
            memory_string = mode_state["memory_string"]
            memory = pkl.loads(base64.b64decode(memory_string))

        conversation_chain = ConversationalRetrievalChain.from_llm(
            llm=self.llm, chain_type="stuff", retriever=self.vectorstore.as_retriever(), memory=memory
        )

        result = conversation_chain.invoke({"question": prompt})

        memory_string = base64.b64encode(pkl.dumps(memory)).decode("utf-8")
        mode_state = {"memory_string": memory_string}
        self.save_mode_state(chat_id, mode_state)
        return result["answer"]
        return result["answer"]
        return result["answer"]



import hashlib
import os

from srai_openai.client_openai_audio import ClientOpenaiAudio
from srai_openai.client_openai_chatgpt import ClientOpenaiChatgpt
from srai_openai.model.prompt_config import PromptConfig

from srai_telegrambot.mode.voice_mode_base import VoiceModeBase


class VoiceModeGpt(VoiceModeBase):

    def __init__(self, system_message_content) -> None:
        super().__init__("text_mode_gpt")
        self.client_chatgtpt = ClientOpenaiChatgpt()
        self.client_audio = ClientOpenaiAudio()
        self.system_message_content = system_message_content

    def _handle_voice(self, chat_id: str, voice_bytes: bytes, voice_bytes_format: str) -> str:
        instance_id = hashlib.sha256(voice_bytes).hexdigest()
        path_file_audio = instance_id + ".ogg"
        with open(path_file_audio, "wb") as f:
            f.write(voice_bytes)
        transcription = self.client_audio.transcription(path_file_audio)
        os.remove(path_file_audio)
        print()
        mode_state = self.try_load_mode_state(chat_id)
        if mode_state is None:
            prompt_config = PromptConfig.create("gpt-4o", self.system_message_content)
        else:
            prompt_config = PromptConfig.from_dict(mode_state)
        prompt_config = prompt_config.append_user_message(transcription["text"])
        prompt_config_result = self.client_chatgtpt.prompt_for_prompt_config(prompt_config)
        self.save_mode_state(chat_id, prompt_config_result.to_dict())
        return prompt_config_result.last_message_text



from srai_core.store.database_memory import DatabaseMemory

from srai_telegrambot.dao_telegram_bot import DaoTelegramBot
from srai_telegrambot.telegram_bot_interface import TelegramBotInterface


class TelegramBotTest(TelegramBotInterface):
    def __init__(self):
        self.dao_telegram_bot = DaoTelegramBot(DatabaseMemory())

    def message_admins(self, text: str):
        print(text)

    def message_chat(self, chat_id: str, text: str):
        print(text)

    def get_dao(self):
        return self.dao_telegram_bot



from abc import ABC, abstractmethod

from telegram import Update
from telegram.ext import CallbackContext

from srai_telegrambot.telegram_bot_interface import TelegramBotInterface


class CommandBase(ABC):
    def __init__(self, command_name: str) -> None:
        self.command_name = command_name
        self.telegram_bot: TelegramBotInterface = None  # type: ignore

    def register(self, telegram_bot: TelegramBotInterface) -> None:
        if telegram_bot is None:
            raise ValueError("telegram_bot cannot be None")
        self.telegram_bot = telegram_bot

    def execute_command_callback(self, update: Update, context: CallbackContext) -> None:
        if self.telegram_bot is None:
            raise ValueError("Cannot run unregistered command")
        # update.message.to_dict()
        # message_id = str(update.message.message_id)
        # chat_id = str(update.message.chat_id)
        # author_id = str(update.message.from_user.id)
        # author_name = update.message.from_user.username
        # message_content = {"message_content_type": "text", "text": update.message.text}
        # message = ChatMessage(message_id, chat_id, author_id, author_name, message_content)
        # self.service_telegram_bot.dao_message.save_message(message)

        self.execute_command(str(update.message.chat_id), update.message.text)

    @abstractmethod
    def execute_command(self, chat_id: str, command_message: str) -> None:
        raise NotImplementedError()



import hashlib
from abc import ABC, abstractmethod
from typing import Optional

from telegram import Update
from telegram.ext import CallbackContext

from srai_telegrambot.telegram_bot_interface import TelegramBotInterface


class TextModeBase(ABC):

    def __init__(self, text_mode_name: str) -> None:
        self.text_mode_name = text_mode_name
        self.telegram_bot: TelegramBotInterface = None  # type: ignore

    def register(self, telegram_bot: TelegramBotInterface) -> None:
        if telegram_bot is None:
            raise ValueError("telegram_bot cannot be None")
        self.telegram_bot = telegram_bot

    def handle_text(self, update: Update, context: CallbackContext):
        if self.telegram_bot is None:
            raise ValueError("Cannot run unregistered text mode")
        message = self._handle_text(str(update.message.chat_id), update.message.text)
        update.message.reply_text(message)

    def try_load_mode_state(self, chat_id: str) -> Optional[dict]:
        state_id: str = self.text_mode_name + "_" + chat_id
        # make sha256 hash of state_id
        state_id_hash: str = hashlib.sha256(state_id.encode()).hexdigest()

        return self.telegram_bot.get_dao().try_load_mode_state(state_id_hash)

    def save_mode_state(self, chat_id: str, mode_state: dict):
        state_id: str = self.text_mode_name + "_" + chat_id
        # make sha256 hash of state_id
        state_id_hash: str = hashlib.sha256(state_id.encode()).hexdigest()
        return self.telegram_bot.get_dao().save_mode_state(state_id_hash, mode_state)

    @abstractmethod
    def _handle_text(self, chat_id: str, text: str) -> str:
        raise NotImplementedError()



from abc import ABC, abstractmethod

from srai_telegrambot.dao_telegram_bot import DaoTelegramBot


class TelegramBotInterface(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def message_admins(self, text: str):
        raise NotImplementedError()

    @abstractmethod
    def message_chat(self, chat_id: str, text: str):
        raise NotImplementedError()

    @abstractmethod
    def get_dao(self) -> DaoTelegramBot:
        raise NotImplementedError()



from srai_telegrambot.command.command_base import CommandBase


class CommandHelp(CommandBase):
    def __init__(self):
        super().__init__("help")

    def execute_command(self, chat_id: str, command_message: str) -> None:
        """Send a message when the command /help is issued."""
        message = "Available commands:\n"
        for command in self.telegram_bot.dict_command.keys():  # type: ignore
            message += f"/{command}\n"
        self.telegram_bot.message_chat(chat_id, message)



from srai_openai.client_openai_chatgpt import ClientOpenaiChatgpt
from srai_openai.model.prompt_config import PromptConfig

from srai_telegrambot.mode.text_mode_base import TextModeBase


class TextModeGpt(TextModeBase):

    def __init__(self, system_message_content: str):
        super().__init__("text_mode_gpt")
        self.client = ClientOpenaiChatgpt()
        self.system_message_content = system_message_content

    def _handle_text(self, chat_id: str, prompt: str) -> str:
        mode_state = self.try_load_mode_state(chat_id)
        if mode_state is None:
            prompt_config = PromptConfig.create("gpt-4o", self.system_message_content)
        else:
            prompt_config = PromptConfig.from_dict(mode_state)
        prompt_config = prompt_config.append_user_message(prompt)
        prompt_config_result = self.client.prompt_for_prompt_config(prompt_config)
        self.save_mode_state(chat_id, prompt_config_result.to_dict())
        return prompt_config_result.last_message_text



from srai_telegrambot.command.command_base import CommandBase


class CommandChatId(CommandBase):
    def __init__(self):
        super().__init__("chat_id")

    def execute_command(self, chat_id: str, command_message: str) -> None:
        """Send a message when the command /help is issued."""
        self.telegram_bot.message_chat(chat_id, f"Chat id of this chat  {chat_id}")
